#include <windows.h>
#include "stdlib.h"
#include "tk.h"
#include "tcl.h"

#define snprintf _snprintf

int
EmbededTkInit(Tcl_Interp *interp)
{
    Tcl_FindExecutable("");

    if (Tcl_Init(interp) == TCL_ERROR) {
	printf("Tcl_Init error: %s\n", Tcl_GetStringResult(interp));
	fflush(stdout);
	return TCL_ERROR;
    }
    if (Tk_Init(interp) == TCL_ERROR) {
	printf("Tk_Init error: %s\n", Tcl_GetStringResult(interp));
	fflush(stdout);
	return TCL_ERROR;
    }

    if (Tcl_Eval(interp, "wm withdraw .") != TCL_OK) {
	printf("Tcl_Eval err %s\n", Tcl_GetStringResult(interp));
	fflush(stdout);
    }
    return TCL_OK;
}

// ------

static int embedtktcount = 0;
static LRESULT CALLBACK tktWndProc(HWND, UINT, WPARAM, LPARAM);

struct tktinfo {
    char *tk_path;
    HWND tkhwnd;
    Tk_Window tktkwin;
    Tcl_Interp *interp;
};

HWND
CreateEmbededTkWindow(Tcl_Interp *interp,
		      DWORD dwStyle, int x, int y, int w, int h,
		      HWND hWndParent, HINSTANCE hInstance,
		      char *retname, size_t retsz)
{
    // this is the name of the tk toplevel embeded widget
    char tkwin_name[1024];
    snprintf(tkwin_name, sizeof(tkwin_name), ".embedtk%d", embedtktcount++);
    strncpy(retname, tkwin_name, retsz);

    static char *wndclassname = "tkt";
    WNDCLASS     wndclass;
    wndclass.style         = 0;
    wndclass.lpfnWndProc   = tktWndProc;
    wndclass.cbClsExtra    = 0;
    wndclass.cbWndExtra    = 0;
    wndclass.hInstance     = hInstance;
    wndclass.hIcon         = LoadIcon (NULL, IDI_APPLICATION);
    wndclass.hCursor       = LoadCursor (NULL, IDC_ARROW);
    wndclass.hbrBackground = (HBRUSH)COLOR_WINDOW;//(HBRUSH) GetStockObject
    wndclass.lpszMenuName  = NULL;
    wndclass.lpszClassName = wndclassname;
    RegisterClass(&wndclass);

    HWND hwnd = CreateWindow(wndclassname, tkwin_name+1, dwStyle, x, y, w, h,
			     hWndParent, NULL, hInstance, NULL);

    tktinfo *tinfo = new tktinfo;
    tinfo->tk_path = strdup(tkwin_name);
    tinfo->interp = interp;
    SetProp(hwnd, "tktinfo", (HANDLE)tinfo);

    char buf[1000];
    snprintf(buf, sizeof(buf), "toplevel %s -use 0x%x", tkwin_name, hwnd);

    // create toplevel window and make it use a diff window id as a base
    if (Tcl_Eval(interp, buf) != TCL_OK) {
	printf("eval err %s\n", Tcl_GetStringResult(interp));
	fflush(stdout);
	return 0;
    }

    // get the tk window token
    {
	Tk_Window tkmain = Tk_MainWindow(interp);
	tinfo->tktkwin = Tk_NameToWindow(interp, (char *)tkwin_name, tkmain);
	RECT topsz;
	GetClientRect(hwnd, &topsz);
	Tk_MoveResizeWindow(tinfo->tktkwin, 0, 0, topsz.right, topsz.bottom);
	Tcl_ServiceAll();
    }

    // get tk child window path
    {
	int i;
	snprintf(buf, sizeof(buf), "winfo id %s", tinfo->tk_path);
	Tcl_Eval(interp, buf);
	Tcl_GetInt(interp, interp->result, &i);
	tinfo->tkhwnd = (HWND)i;
    }

    return hwnd;
}

static LRESULT CALLBACK
tktWndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    Tcl_SetServiceMode(TCL_SERVICE_NONE);

    tktinfo *tinfo = (tktinfo*)GetProp(hwnd, "tktinfo");

    switch (message) {
	case WM_DESTROY:
	    {
		char buf[2000];
		snprintf(buf, sizeof(buf), "destroy %s", tinfo->tk_path);
		Tcl_Eval(tinfo->interp, buf);
		Tcl_ServiceAll();
		free(tinfo->tk_path);
		delete tinfo;
	    }
	    break;

	case WM_GETTEXT:
	    strncpy((char*)lParam, tinfo->tk_path, wParam);
	    return 0;

	case WM_KEYDOWN:
	case WM_CHAR:
	case WM_KEYUP:
	case WM_SETFOCUS:
	case WM_MOUSEMOVE:
	    if (tinfo) ::SendMessage(tinfo->tkhwnd, message, wParam, lParam);
	    break;

	case WM_SIZE:
	    {
		if (tinfo) {
		    RECT topsz;
		    GetClientRect(hwnd, &topsz);
		    Tk_MoveResizeWindow(tinfo->tktkwin, 0, 0,
					topsz.right, topsz.bottom);
		    char buf[2000];
		    snprintf(buf, sizeof(buf), "update idletasks");
		    Tcl_Eval(tinfo->interp, buf);
		    Tcl_ServiceAll();
		}
	    }
    }
    Tcl_SetServiceMode(TCL_SERVICE_ALL);
    return DefWindowProc(hwnd, message, wParam, lParam);
}
